@@ -1,7 +1,9 @@
 from pathlib import Path
+from typing import Literal
 import typer
 import click
-from snowcli.snow_connector import connect_to_snowflake
+import os
+from ..utils import get_snowpark_session
 from ..plugin_uploader import PluginUploader
 from ..native_app_packaging import NativeAppPackaging
 from ..initialiser import new_plugin_generator
@@ -72,11 +74,26 @@ def upload(
         "--partner_plugin",
         help="Marks the plugin as being partner-developed, which excludes it from BYO billing behaviour. Billing events will be created by both the Sync Engine and the plugin (if defined)"
     ),
+    log_level: str = typer.Option(
+        "INFO",
+        "--log_level",
+        help="The log level to set in the native app manifest. One of TRACE,DEBUG,INFO,WARN,ERROR,FATAL,OFF",
+    ),
+    udf_directory_override: Path = typer.Option(
+        None,
+        "--udf_directory_override",
+        help="The directory containing UDFs to upload to the plugin schema. If not provided, the whole source folder will be scanned",
+    ),
+    distribution: str = typer.Option(
+        "INTERNAL",
+        "--distribution",
+        help="Distribution of the application package. One of INTERNAL, EXTERNAL",
+    ),
 ):
     """
     Packages a plugin directory, uploads to a Snowflake stage and creates an Application Package.
     """
-    snowflake_connection = connect_to_snowflake(connection_name=environment)
+    snowflake_connection = get_snowpark_session(connection_name=environment)
     uploader = PluginUploader(snowflake_connection=snowflake_connection)
     target_schema = uploader.upload_plugin(
         plugin_directory=plugin_path.resolve(),
@@ -84,7 +101,9 @@ def upload(
         schema_name=target_schema,
         stage_name=target_stage,
         is_airbyte=airbyte,
-        partner_plugin=partner_plugin
+        partner_plugin=partner_plugin,
+        log_level=log_level,
+        udf_directory_override=udf_directory_override,
     )
     if package_name is None:
         package_name = target_schema + "_PLUGIN"
@@ -92,7 +111,10 @@ def upload(
         snowflake_connection=snowflake_connection,
         package_name=package_name,
     )
-    packager.create_package(developer_role=developer_role)
+    if distribution not in ["INTERNAL", "EXTERNAL"]:
+        print("Distribution must be one of INTERNAL or EXTERNAL")
+        raise typer.Abort()
+    packager.create_package(developer_role=developer_role,distribution=distribution)
     patch_number = packager.create_package_version(
         code_database=target_database,
         code_schema=target_schema,
@@ -131,7 +153,7 @@ def deploy(
     """
     Creates a plugin Application from an uploaded Application package.
     """
-    snowflake_connection = connect_to_snowflake(connection_name=environment)
+    snowflake_connection = get_snowpark_session(connection_name=environment)
     packager = NativeAppPackaging(
         snowflake_connection=snowflake_connection,
         package_name=package_name,
@@ -175,7 +197,7 @@ def register(
     """
     Registers a deployed plugin Application with the Omnata Application.
     """
-    snowflake_connection = connect_to_snowflake(connection_name=environment)
+    snowflake_connection = get_snowpark_session(connection_name=environment)
     registerer = PluginRegistration(snowflake_connection=snowflake_connection)
     registerer.register_plugin(
         plugin_application_name=plugin_application_name,
