@@ -27,6 +27,7 @@ class NativeAppPackaging:
         self,
         developer_role: Optional[str] = None,
         distribution: Literal["INTERNAL", "EXTERNAL"] = "INTERNAL",
+        enable_release_channels: bool = False,
     ) -> bool:
         """
         Creates a native application package in Snowflake.
@@ -39,7 +40,8 @@ class NativeAppPackaging:
         if len(app_packages) == 0:
             self.session.sql(
                 f"""CREATE APPLICATION PACKAGE {self.package_name} 
-                                        DISTRIBUTION={distribution}"""
+                                        DISTRIBUTION={distribution}
+                                        ENABLE_RELEASE_CHANNELS={str(enable_release_channels).upper()}"""
             ).collect()
         if developer_role is not None:
             self.session.sql(
@@ -80,6 +82,7 @@ class NativeAppPackaging:
         version_name: str,
         patch_number: Optional[int] = None,
         debug_mode: bool = True,
+        authorize_telemetry_event_sharing: bool = True,
     ) -> bool:
         """
         Creates a native application in Snowflake.
@@ -106,7 +109,8 @@ class NativeAppPackaging:
                 f"""CREATE APPLICATION {application_name} 
                                         FROM APPLICATION PACKAGE {self.package_name}
                                         USING VERSION {version_name} patch {patch_number}
-                                        DEBUG_MODE={str(debug_mode).upper()}"""
+                                        DEBUG_MODE={str(debug_mode).upper()}
+                                        AUTHORIZE_TELEMETRY_EVENT_SHARING={str(authorize_telemetry_event_sharing).upper()}"""
             ).collect()
             return True
         self.session.sql(
